/* hacker port driver for Linux
 * version 0.1 date 2004-05-05
 * part of the hacker port project - http[s]://1stein.schuermans.info/hackerport/
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/wrapper.h>
#include <linux/slab.h>
#include <linux/usb.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/atomic.h>

#include "hkp_drv.h"



//#define DEBUG_USB_INT
//#define DEBUG_USB_BULK
//#define DEBUG_MEM
#ifndef DEBUG
#undef DEBUG_USB_INT
#undef DEBUG_USB_BULK
#undef DEBUG_MEM
#endif



//version information
#define hkp_drv_name "hacker port driver for Linux"
#define hkp_drv_ver "version 0.1 date 2004-05-05"


//number of elements in an array
#define count( array ) (sizeof( (array) ) / sizeof( (array)[0] ))



//info macros
#define hkp_drv_info( text ) printk( KERN_INFO "hkp_drv: "text"\n" )
#define hkp_drv_infof( format, arg... ) printk( KERN_INFO "hkp_drv: "format"\n", ##arg )
//warning macros
#define hkp_drv_warning( text ) printk( KERN_WARNING "hkp_drv: "text"\n" )
#define hkp_drv_warningf( format, arg... ) printk( KERN_WARNING "hkp_drv: "format"\n", ##arg )
//error macros
#define hkp_drv_error( text ) printk( KERN_ERR "hkp_drv: "text"\n" )
#define hkp_drv_errorf( format, arg... ) printk( KERN_ERR "hkp_drv: "format"\n", ##arg )
//debug macros
#ifdef DEBUG
#define hkp_drv_debug( text ) printk( KERN_DEBUG "hkp_drv: "__FUNCTION__": "text"\n" )
#define hkp_drv_debugf( format, arg... ) printk( KERN_DEBUG "hkp_drv: "__FUNCTION__": "format"\n", ##arg )
#else
#define hkp_drv_debug( text )
#define hkp_drv_debugf( format, arg... )
#endif



//global constants concerning the device
//(we only use one vendor id product id pair and identify the devices using the high-byte of the device-revision)
#define hkp_drv_id_vendor 0x6666 //vendor id: prototype device
#define hkp_drv_id_product 0x3353 //product id: '3S' for Stefan Schuermans Software
#define hkp_drv_bcd_device_high 0x03 //high-byte of device-revision: 0x03=HackerPort
//global constants concerning the module
#define hkp_drv_minor_base 0xB0 //base of minor device number
#define hkp_drv_minor_range 0x10 //range of minor device numbers
#define hkp_drv_devfs_node_name "hkp_drv" //node name for devfs



#ifdef DEBUG_MEM
//memory debug
struct semaphore my_alloc_lock;
int my_alloc_cnt = 0;
void * my_kmalloc( int len, int flags )
{
  void * ptr;
  down( &my_alloc_lock );
  ptr = kmalloc( len, flags );
  if( ptr )
  {
    my_alloc_cnt++;
    printk( KERN_INFO "ptr=%p alloc_cnt=%d\n", ptr, my_alloc_cnt );
  }
  up( &my_alloc_lock );
  return ptr;
}
void my_kfree( void * ptr )
{
  down( &my_alloc_lock );
  kfree( ptr );
  my_alloc_cnt--;
  printk( KERN_INFO "ptr=%p alloc_cnt=%d\n", ptr, my_alloc_cnt );
  up( &my_alloc_lock );
}
#define kmalloc my_kmalloc
#define kfree my_kfree
#endif



//external variables
extern devfs_handle_t usb_devfs_handle; //handle of global usb directory of devfs



//type for private data of a file
typedef struct t_hkp_drv_file_priv
{
  struct t_hkp_drv_dev * p_dev; //pointer to associated device
  struct t_hkp_drv_file_priv * p_next, * p_prev; //pointer to next and previous file that have opened same device

  wait_queue_head_t poll_wait_queue; //waitqueue for poll-call
} st_hkp_drv_file_priv;

//type for private data of a device
typedef struct t_hkp_drv_dev
{
  struct semaphore file_lock; //used to syncronize file operations on this device
  struct semaphore lock; //used to lock this structure

  struct usb_device * p_usb_dev; //saved parameters of probe function
  unsigned int interface_no;
  const struct usb_device_id * p_usb_dev_id;

  unsigned char dev_no_sub_minor; //subminor device number (minor relative to base)
  unsigned char dev_no_minor; //minor device number

  unsigned char int_out_endpoint_addr; //address of the interrupt out endpoint
  int int_out_max_packet_size; //maximum size of packets for interrupt out endpoint
  int int_out_interval; //interval of interrupt out endpoint
  unsigned char * p_int_out_buffer; //pointer to buffer for interrupt out
  struct urb * p_int_out_urb; //interrupt out URB
  int int_out_urb_linked; //boolean flag, if interrupt out URB is submitted/linked

  unsigned char int_in_endpoint_addr; //address of the interrupt in endpoint
  int int_in_max_packet_size; //maximum size of packets for interrupt in endpoint
  int int_in_interval; //interval of interrupt in endpoint
  unsigned char * p_int_in_buffer; //pointer to buffer for interrupt in
  struct urb * p_int_in_urb; //interrrupt in URB
  int int_in_urb_linked; //boolean flag, if interrupt in URB is submitted/linked

  unsigned char bulk_out_endpoint_addr; //address of the bulk out endpoint
  int bulk_out_max_packet_size; //maximum size of packets for bulk out endpoint
  unsigned char * p_bulk_out_buffer; //pointer to buffer for bulk out
  struct urb * p_bulk_out_urb; //bulk out URB
  int bulk_out_urb_linked; //boolean flag, if bulk out URB is submitted/linked

  unsigned char bulk_in_endpoint_addr; //address of the bulk in endpoint
  int bulk_in_max_packet_size; //maximum size of packets for bulk in endpoint
  unsigned char * p_bulk_in_buffer; //pointer to buffer for bulk in
  struct urb * p_bulk_in_urb; //bulk in URB
  int bulk_in_urb_linked; //boolean flag, if bulk in URB is submitted/linked

  char devfs_node_name[64]; //name of devfs node
  devfs_handle_t h_devfs_node; //handle of devfs node

  int open_cnt; //numer of times this file is opened at the moment
  struct t_hkp_drv_file_priv file_priv_list; //anchor of list with files that have opened this device

  unsigned char out_vals[2]; //values for output ports (0..255)
  unsigned short pwm_vals[2]; //values for PWM outputs (0..1000)

  unsigned char in_vals[2]; //values of input ports (0..255)
  unsigned short ana_vals[5]; //values of analog inputs (0..1023)
} st_hkp_drv_dev;



//table with pointers to private device data of devices
//(indexed by subminor)
static struct t_hkp_drv_dev * hkp_drv_dev_table[hkp_drv_minor_range];
//lock for this table
static struct semaphore hkp_drv_dev_table_lock;

//prototype of task function (needed by declaration of task queue)
static void hkp_drv_task( void * p_ingnored );
//taskqueue to do some things that cannot be done in interrupt
static struct tq_struct hkp_drv_task_queue =
{
  list: { NULL },
  sync: 0, 
  routine: hkp_drv_task,
  data: NULL
};
//boolean flags for tasks to do in next hkp_drv_task, modified by atomic bitops
//(hkp_drv_task is task for things that cannot be done in interrupt)
static long hkp_drv_task_flags[hkp_drv_minor_range];
#define hkp_drv_task_flag_bulk_out_try 0 //try to send something using bulk output



//convert character to long looking like binary when printed in hexadecimal
static unsigned long bin( unsigned char val )
{
  return (val & 0x01 ? 0x00000001 : 0)
       | (val & 0x02 ? 0x00000010 : 0)
       | (val & 0x04 ? 0x00000100 : 0)
       | (val & 0x08 ? 0x00001000 : 0)
       | (val & 0x10 ? 0x00010000 : 0)
       | (val & 0x20 ? 0x00100000 : 0)
       | (val & 0x40 ? 0x01000000 : 0)
       | (val & 0x80 ? 0x10000000 : 0);
}



//*** interrupt output interrupt functions

//callback-procedure for USB interrupt output
//(interrupt context)
static void hkp_drv_int_out_callback( struct urb * p_urb )
{
  struct t_hkp_drv_dev * p_dev;
#ifdef DEBUG_USB_INT
  int i, j;
  char dbg_text[256];
#endif

  //get device
  p_dev = (struct t_hkp_drv_dev *)p_urb->context;

  //error
  if( p_urb->status != USB_ST_NOERROR )
    //print error message
    hkp_drv_errorf( "interrupt output reported error: %d", p_urb->status );
  //no error
  else
  {
    //not all data was sent
    if( p_urb->actual_length != p_urb->transfer_buffer_length )
      //print warning message
      hkp_drv_warningf( "interrupt output returned a different length: %d (expected: %d)",
          p_urb->actual_length, p_urb->transfer_buffer_length );
#ifdef DEBUG_USB_INT
    //print interrupt message with data
    dbg_text[0] = 0;
    for( i = 0, j = 0; i < p_urb->actual_length && j < sizeof( dbg_text ) - 4; i++, j += 3 )
      sprintf( dbg_text + j, "%02X ", ((unsigned char *)p_urb->transfer_buffer)[i] );
    hkp_drv_debugf( "interrupt output done (len=%d): %s", p_urb->actual_length, dbg_text );
#endif
  }

  //update data for output ports
  //(can only be done here, because interrupt output buffer is in use all the time)
  //FIXME: this might happen in the middle of writing these values
  if( p_dev->int_out_max_packet_size >= 1 )
    p_dev->p_int_out_buffer[0] = p_dev->out_vals[0];
  if( p_dev->int_out_max_packet_size >= 2 )
    p_dev->p_int_out_buffer[1] = p_dev->out_vals[1];
  
  //update data for PWM outputs
  //(can only be done here, because interrupt output buffer is in use all the time)
  //FIXME: this might happen in the middle of writing these values
  if( p_dev->int_out_max_packet_size >= 4 )
    *(__u16 *)&p_dev->p_int_out_buffer[2] = __cpu_to_be16( (p_dev->pwm_vals[0] & 0x3FF) << 6 );
  if( p_dev->int_out_max_packet_size >= 6 )
    *(__u16 *)&p_dev->p_int_out_buffer[4] = __cpu_to_be16( (p_dev->pwm_vals[1] & 0x3FF) << 6 );
}



//*** interrupt input interrupt functions

//process a interrupt in message
//(may be called in interrupt context with device locked by someone else)
static void hkp_drv_int_in_msg( struct t_hkp_drv_dev * p_dev, unsigned char * p_msg, int msg_len )
{
  //check minimum langth of message
  if( msg_len < 12 )
    return;

  //save new values of input ports
  //FIXME: this might happen in the middle of reading these values
  p_dev->in_vals[0] = p_msg[0];
  p_dev->in_vals[1] = p_msg[1];

  //save new values of analog inputs
  //FIXME: this might happen in the middle of reading these values
  p_dev->ana_vals[0] = __be16_to_cpu( *(__u16 *)&p_msg[2] ) >> 6;
  p_dev->ana_vals[1] = __be16_to_cpu( *(__u16 *)&p_msg[4] ) >> 6;
  p_dev->ana_vals[2] = __be16_to_cpu( *(__u16 *)&p_msg[6] ) >> 6;
  p_dev->ana_vals[3] = __be16_to_cpu( *(__u16 *)&p_msg[8] ) >> 6;
  p_dev->ana_vals[4] = __be16_to_cpu( *(__u16 *)&p_msg[10] ) >> 6;
}

//callback-procedure for USB interrupt input
//(interrupt context)
static void hkp_drv_int_in_callback( struct urb * p_urb )
{
  struct t_hkp_drv_dev * p_dev;
#ifdef DEBUG_USB_INT
  int i, j;
  char dbg_text[256];
#endif

  //get device
  p_dev = (struct t_hkp_drv_dev *)p_urb->context;

  //error
  if( p_urb->status != USB_ST_NOERROR )
    //print error message
    hkp_drv_errorf( "interrupt input reported error: %d", p_urb->status );
  //no error
  else
  {
#ifdef DEBUG_USB_INT
    //print interrupt message with data
    dbg_text[0] = 0;
    for( i = 0, j = 0; i < p_urb->actual_length && j < sizeof( dbg_text ) - 4; i++, j += 3 )
      sprintf( dbg_text + j, "%02X ", ((unsigned char *)p_urb->transfer_buffer)[i] );
    hkp_drv_debugf( "interrupt input done (len=%d): %s", p_urb->actual_length, dbg_text );
#endif

    //process the message
    //(device may be locked by someone else, we are in interrupt context)
    hkp_drv_int_in_msg( p_dev, (unsigned char *)p_urb->transfer_buffer, p_urb->actual_length );
  }
}



//*** bulk output interrupt functions

//callback-procedure for USB bulk output
//(interrupt context)
static void hkp_drv_bulk_out_callback( struct urb * p_urb )
{
  struct t_hkp_drv_dev * p_dev;
#ifdef DEBUG_USB_BULK
  int i, j;
  char dbg_text[256];
#endif

  //get device
  p_dev = (struct t_hkp_drv_dev *)p_urb->context;

  //bulk out URB is no more submitted/linked
  p_dev->bulk_out_urb_linked = 0;

  //error
  if( p_urb->status != USB_ST_NOERROR )
    //print error message
    hkp_drv_errorf( "bulk output reported error: %d", p_urb->status );
  //no error
  else
  {
    //not all data was sent
    if( p_urb->actual_length != p_urb->transfer_buffer_length )
      //print warning message
      hkp_drv_warningf( "bulk output returned a different length: %d (expected: %d)",
          p_urb->actual_length, p_urb->transfer_buffer_length );
#ifdef DEBUG_USB_BULK
    //print bulk message with data
    dbg_text[0] = 0;
    for( i = 0, j = 0; i < p_urb->actual_length && j < sizeof( dbg_text ) - 4; i++, j += 3 )
      sprintf( dbg_text + j, "%02X ", ((unsigned char *)p_urb->transfer_buffer)[i] );
    hkp_drv_debugf( "bulk output done (len=%d): %s", p_urb->actual_length, dbg_text );
#endif
  }

  //set flag for execution of hkp_drv_bulk_out_try for this device
  //(use test_and_x here to be smp safe)
  test_and_set_bit( hkp_drv_task_flag_bulk_out_try, &hkp_drv_task_flags[p_dev->dev_no_sub_minor] );
  //schedule task for things that can not be done in interrupt
  schedule_task( &hkp_drv_task_queue );
}



//*** bulk input interrupt functions

//process a bulk in message
//(may be called in interrupt context with device locked by someone else)
static void hkp_drv_bulk_in_msg( struct t_hkp_drv_dev * p_dev, unsigned char * p_msg, int msg_len )
{
}

//callback-procedure for USB bulk input
//(interrupt context)
static void hkp_drv_bulk_in_callback( struct urb * p_urb )
{
  struct t_hkp_drv_dev * p_dev;
#ifdef DEBUG_USB_BULK
  int i, j;
  char dbg_text[256];
#endif

  //get device
  p_dev = (struct t_hkp_drv_dev *)p_urb->context;

  //bulk in URB is no more submitted/linked
  p_dev->bulk_in_urb_linked = 0;

  //error
  if( p_urb->status != USB_ST_NOERROR )
    //print error message
    hkp_drv_errorf( "bulk input reported error: %d", p_urb->status );
  //no error
  else
  {
#ifdef DEBUG_USB_BULK
    //print bulk message with data
    dbg_text[0] = 0;
    for( i = 0, j = 0; i < p_urb->actual_length && j < sizeof( dbg_text ) - 4; i++, j += 3 )
      sprintf( dbg_text + j, "%02X ", ((unsigned char *)p_urb->transfer_buffer)[i] );
    hkp_drv_debugf( "bulk input done (len=%d): %s", p_urb->actual_length, dbg_text );
#endif

    //process the message
    //(device may be locked by someone else, we are in interrupt context)
    hkp_drv_bulk_in_msg( p_dev, (unsigned char *)p_urb->transfer_buffer, p_urb->actual_length );

    //set up data in bulk in URB
    FILL_BULK_URB( p_dev->p_bulk_in_urb, p_dev->p_usb_dev, usb_rcvbulkpipe( p_dev->p_usb_dev, p_dev->bulk_in_endpoint_addr ),
        p_dev->p_bulk_in_buffer, p_dev->bulk_in_max_packet_size, hkp_drv_bulk_in_callback,
        p_dev );

    //bulk in URB will be linked/submitted soon
    //(do this here because callback could perhaps be called from usb_submit_urb)
    p_dev->bulk_in_urb_linked = 1;
    //resubmit bulk in URB
    if( usb_submit_urb( p_dev->p_bulk_in_urb ) )
    {
      //bulk in URB is not linked/submitted
      p_dev->bulk_in_urb_linked = 0;
      //print error message
      hkp_drv_error( "could not submit bulk in URB" );
    }
    else
    {
      //print info that bulk in was started
#ifdef DEBUG_USB_BULK
      hkp_drv_debugf( "bulk input started (max_len=%d)", p_dev->p_bulk_in_urb->transfer_buffer_length );
#endif
    }
  }
}



//*** functions to do some things that cannot be done in interrupt

//try to send something using bulk output
//(must be called with device locked)
static void hkp_drv_bulk_out_try( struct t_hkp_drv_dev * p_dev )
{
}

//task to do some things that cannot be done in interrupt
static void hkp_drv_task( void * p_ingnored )
{
  int i;
  struct t_hkp_drv_dev * p_dev;

  //before waiting for any lock, peek at flags to see if we have got something to do
  for( i = 0; i < hkp_drv_minor_range; i++ )
  {
    if( test_bit( hkp_drv_task_flag_bulk_out_try, &hkp_drv_task_flags[i] ) )
      break;
  }
  //nothing to do
  if( i >= hkp_drv_minor_range )
    return;

  //lock device table
  down( &hkp_drv_dev_table_lock );

  //check all devices in device table
  for( i = 0; i < hkp_drv_minor_range; i++ )
  {
    //get pointer to device
    p_dev = hkp_drv_dev_table[i];
    //device is there
    if( p_dev != NULL )
    {
      //lock the device
      down( &p_dev->lock );

      //call functions whose flag is set and clear the flag
      if( test_and_clear_bit( hkp_drv_task_flag_bulk_out_try, &hkp_drv_task_flags[i] ) )
        hkp_drv_bulk_out_try( p_dev );

      //unlock the device
      up( &p_dev->lock );
    }
  }

  //unlock device table
  up( &hkp_drv_dev_table_lock );
}



//*** support functions for non-interrupt-time - device locked

//free a device and everything that belongs to it
//device is also removed from the device table
//(the device table must be locked)
static void hkp_drv_dev_free( struct t_hkp_drv_dev * p_dev )
{
  //remove device from the device table
  hkp_drv_dev_table[p_dev->dev_no_sub_minor] = NULL;

  //free URBs
  if( p_dev->p_int_out_urb != NULL )
  {
    if( p_dev->int_out_urb_linked )
      usb_unlink_urb( p_dev->p_int_out_urb );
    usb_free_urb( p_dev->p_int_out_urb );
  }
  if( p_dev->p_int_in_urb != NULL )
  {
    if( p_dev->int_in_urb_linked )
      usb_unlink_urb( p_dev->p_int_in_urb );
    usb_free_urb( p_dev->p_int_in_urb );
  }
  if( p_dev->p_bulk_out_urb != NULL )
  {
    if( p_dev->bulk_out_urb_linked )
      usb_unlink_urb( p_dev->p_bulk_out_urb );
    usb_free_urb( p_dev->p_bulk_out_urb );
  }
  if( p_dev->p_bulk_in_urb != NULL )
  {
    if( p_dev->bulk_in_urb_linked )
      usb_unlink_urb( p_dev->p_bulk_in_urb );
    usb_free_urb( p_dev->p_bulk_in_urb );
  }
  //free buffers
  if( p_dev->p_int_out_buffer != NULL )
    kfree( p_dev->p_int_out_buffer );
  if( p_dev->p_int_in_buffer != NULL )
    kfree( p_dev->p_int_in_buffer );
  if( p_dev->p_bulk_out_buffer != NULL )
    kfree( p_dev->p_bulk_out_buffer );
  if( p_dev->p_bulk_in_buffer != NULL )
    kfree( p_dev->p_bulk_in_buffer );

  //free device structure
  kfree( p_dev );
}



//*** file operations

//an userspace process is reading from one of the devices
static ssize_t hkp_drv_file_read( struct file * p_file, char * p_buffer, size_t count, loff_t * p_pos )
{
  struct t_hkp_drv_file_priv * p_file_priv;
  struct t_hkp_drv_dev * p_dev;
  char * p_str;
  int len, start, cnt;

  //test pointer
  if( verify_area( VERIFY_WRITE, p_buffer, count ) )
    //pointer not okay
    return -EFAULT;

  //get private file
  p_file_priv = (struct t_hkp_drv_file_priv *)p_file->private_data;
  //get private device
  p_dev = p_file_priv->p_dev;
  //no pointer to device
  if( p_dev == NULL )
    //so there is no device
    return -ENODEV;

  //lock device for file operations
  down( &p_dev->file_lock );
  //lock device
  down( &p_dev->lock );

  //allocate buffer for return string
  p_str = (char *)kmalloc( 512, GFP_KERNEL );

  //assemble return string
  len = sprintf( p_str,
                 "input A: %3d = 0x%02X = 0b%08lX\n"
                 "input B: %3d = 0x%02X = 0b%08lX\n"
                 "analog input A: %4d (%4ld mV)\n"
                 "analog input B: %4d (%4ld mV)\n"
                 "analog input C: %4d (%4ld mV)\n"
                 "analog input D: %4d (%4ld mV)\n"
                 "analog input E: %4d (%4ld mV)\n",
                 p_dev->in_vals[0], p_dev->in_vals[0], bin( p_dev->in_vals[0] ),
                 p_dev->in_vals[1], p_dev->in_vals[1], bin( p_dev->in_vals[1] ),
                 p_dev->ana_vals[0], (unsigned long)p_dev->ana_vals[0] * 5000 / 1023,
                 p_dev->ana_vals[1], (unsigned long)p_dev->ana_vals[1] * 5000 / 1023,
                 p_dev->ana_vals[2], (unsigned long)p_dev->ana_vals[2] * 5000 / 1023,
                 p_dev->ana_vals[3], (unsigned long)p_dev->ana_vals[3] * 5000 / 1023,
                 p_dev->ana_vals[4], (unsigned long)p_dev->ana_vals[4] * 5000 / 1023 );

  //get position and length of data to return
  start = *p_pos < 512 ? (int)*p_pos : 512;
  cnt = min( (int)(len > start ? len - start : 0), (int)count );
  //advance file pointer
  *p_pos += cnt;

  //copy data to user space
  copy_to_user( p_buffer, p_str + start, cnt );

  //free buffer for return string
  kfree( p_str );

  //debug message
  hkp_drv_debugf( "read from file %08X (%d, %d): %d bytes", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, cnt );

  //unlock device
  up( &p_dev->lock );
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return number of bytes read
  return cnt;
}

//an userspace process is writing to one of the devices
static ssize_t hkp_drv_file_write( struct file * p_file, const char * p_buffer, size_t count, loff_t * p_pos )
{
  struct t_hkp_drv_file_priv * p_file_priv;
  struct t_hkp_drv_dev * p_dev;
  char * p_chr, * p_line;
  int pos, len, end, line_len, cnt, val;

  //test pointer
  if( verify_area( VERIFY_READ, p_buffer, count ) )
    //pointer not okay
    return -EFAULT;

  //get private file
  p_file_priv = (struct t_hkp_drv_file_priv *)p_file->private_data;
  //get private device
  p_dev = p_file_priv->p_dev;
  //no pointer to device
  if( p_dev == NULL )
    //so there is no device
    return -ENODEV;

  //lock device for file operations
  down( &p_dev->file_lock );
  //lock device
  down( &p_dev->lock );

  //parse commands
  for( pos = 0, len = (int)count; ; )
  {
    //get next line
    p_chr = memchr( p_buffer + pos, '\n', len );
    if( p_chr == NULL )
      break;
    *p_chr = 0;
    end = p_chr - p_buffer;
    line_len = end - pos;
    p_line = (char *)(p_buffer + pos);

    //commands
    if( sscanf( p_line, "output A: %i", &val ) == 1 && val >= 0 && val <= 255 )
    {
      p_dev->out_vals[0] = (unsigned char)val;
      hkp_drv_debugf( "command: output A: %3d = 0x%02X = 0b%08lX",
                      val, val, bin( (unsigned char)val ) );
    }
    if( sscanf( p_line, "output B: %i", &val ) == 1 && val >= 0 && val <= 255 )
    {
      p_dev->out_vals[1] = (unsigned char)val;
      hkp_drv_debugf( "command: output B: %3d = 0x%02X = 0b%08lX",
                      val, val, bin( (unsigned char)val ) );
    }
    if( sscanf( p_line, "PWM output A: %i", &val ) == 1 && val >= 0 && val <= 1000 )
    {
      p_dev->pwm_vals[0] = (unsigned short)val;
      hkp_drv_debugf( "command: PWM output A: %4d (%3d.%1d%%)\n",
                      val, val / 10, val % 10 );
    }
    if( sscanf( p_line, "PWM output B: %i", &val ) == 1 && val >= 0 && val <= 1000 )
    {
      p_dev->pwm_vals[1] = (unsigned short)val;
      hkp_drv_debugf( "command: PWM output B: %4d (%3d.%1d%%)\n",
                      val, val / 10, val % 10 );
    }

    //go to end of current line
    pos += line_len;
    len -= line_len;
  }
  cnt = (int)count;

  //debug message
  hkp_drv_debugf( "wrote to file %08X (%d, %d): %d bytes", (int)p_file, USB_MAJOR, p_dev->dev_no_minor, cnt );

  //unlock device
  up( &p_dev->lock );
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return number of bytes written to this device
  return cnt;
}

//an userspace process is opening one of the devices
static int hkp_drv_file_open( struct inode * p_inode, struct file * p_file )
{
  struct t_hkp_drv_file_priv * p_file_priv;
  struct t_hkp_drv_dev * p_dev;
  int sub_minor;
  
  //get subminor device number
  sub_minor = MINOR( p_inode->i_rdev ) - hkp_drv_minor_base;
  if( sub_minor < 0  || sub_minor >= hkp_drv_minor_range )
    //device not one of our devices
    return -ENODEV;

  //lock device table
  down( &hkp_drv_dev_table_lock );
  //get device structure for this minor
  p_dev = hkp_drv_dev_table[sub_minor];
  //device not attached
  if( p_dev == NULL )
  {
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //device not available
    return -ENODEV;
  }

  //device was unplugged
  if( p_dev->p_usb_dev == NULL )
  {
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //device not available
    return -ENODEV;
  }

  //create private file data
  p_file_priv = (struct t_hkp_drv_file_priv *)kmalloc( sizeof( struct t_hkp_drv_file_priv ), GFP_KERNEL );
  if( p_file_priv == NULL )
  {
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //no memory available
    return -ENOMEM;
  }

  //initialize private file data
  //(p_dev and p_next_file_priv are initialized a few lines down)
  //initialize waitqueue for poll-call
  init_waitqueue_head( &p_file_priv->poll_wait_queue );
 
  //lock device for file operations
  down( &p_dev->file_lock );
  //lock device
  down( &p_dev->lock );

  //unlock device table
  up( &hkp_drv_dev_table_lock );

  //increment our usage count
  p_dev->open_cnt++;

  //save pointer to private file data in file
  p_file->private_data = (void *)p_file_priv;
  //save pointer to device in the file's private structure
  p_file_priv->p_dev = p_dev;
  //put private file data into list in device structure
  //the list stays consistent all the time
  //(reading in interrupt is allowed when using only p_next or only p_prev)
  p_file_priv->p_prev = p_dev->file_priv_list.p_prev;
  p_file_priv->p_next = &p_dev->file_priv_list;
  p_file_priv->p_next->p_prev = p_file_priv;
  p_file_priv->p_prev->p_next = p_file_priv;

  //debug message
  hkp_drv_debugf( "opened file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //unlock device
  up( &p_dev->lock );
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return success
  return 0;
}

//an userspace process is closing one of the devices
static int hkp_drv_file_release( struct inode * p_inode, struct file * p_file )
{
  struct t_hkp_drv_file_priv * p_file_priv;
  struct t_hkp_drv_dev * p_dev;

  //get private file
  p_file_priv = (struct t_hkp_drv_file_priv *)p_file->private_data;
  //get private device
  p_dev = p_file_priv->p_dev;
  //no pointer to device
  if( p_dev == NULL )
    //so there is no device
    return -ENODEV;

  //lock device table
  down( &hkp_drv_dev_table_lock );

  //lock device for file operations
  down( &p_dev->file_lock );
  //lock device
  down( &p_dev->lock );

  //wake up poll-waitqueue of file (in case that another thread is doing a poll)
  wake_up_interruptible( &p_file_priv->poll_wait_queue );

  //remove private file data from list in device structure
  //the list stays consistent all the time
  //(reading in interrupt is allowed when using only p_next or only p_prev)
  p_file_priv->p_next->p_prev = p_file_priv->p_prev;
  p_file_priv->p_prev->p_next = p_file_priv->p_next;

  //free private file data
  kfree( p_file_priv );

  //check if device is open
  if( p_dev->open_cnt <= 0 )
  {
    //unlock device
    up( &p_dev->lock );
    //unlock device for file operations
    up( &p_dev->file_lock );
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //there is no such device
    return -ENODEV;
  }

  //debug message
  hkp_drv_debugf( "closed file %08X (%d, %d)", (int)p_file, USB_MAJOR, p_dev->dev_no_minor );

  //decrement usage count
  p_dev->open_cnt--;

  //device was unplugged
  if( p_dev->p_usb_dev == NULL )
  {
    //device is opened by someone else
    if( p_dev->open_cnt > 0 )
    {
      //unlock the device
      up( &p_dev->lock );
      //unlock device for file operations
      up( &p_dev->file_lock );
    }
    //device is not opened by anyone else
    else
    {
      //unlock the device
      up( &p_dev->lock );
      //unlock device for file operations
      up( &p_dev->file_lock );
      //free the device (also frees buffers and URBs)
      //(and remove it from the device table)
      hkp_drv_dev_free( p_dev );
    }
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //success
    return 0;
  }

  //unlock device
  up( &p_dev->lock );
  //unlock device table
  up( &hkp_drv_dev_table_lock );
  //unlock device for file operations
  up( &p_dev->file_lock );

  //return success
  return 0;
}

//structure for supported file operations
static struct file_operations hkp_drv_file_operations =
{
  owner:   THIS_MODULE, //because of this, we don't need the MOD_INC/DEC_USE_COUNT macros
  llseek:  NULL,
  read:    hkp_drv_file_read,
  write:   hkp_drv_file_write,
  readdir: NULL,
  poll:    NULL,
  ioctl:   NULL,
  mmap:    NULL,
  open:    hkp_drv_file_open,
  flush:   NULL,
  release: hkp_drv_file_release,
};



//*** USB device operations

//device is attached to USB
static void * hkp_drv_probe( struct usb_device * p_usb_dev, unsigned int interface_no, const struct usb_device_id * p_usb_dev_id )
{
  unsigned char sub_minor;
  struct t_hkp_drv_dev * p_dev;
  struct usb_interface * p_interface;
  struct usb_interface_descriptor * p_interface_descriptor;
  int i, endpoints_ok;

  //check if vendor id and product id match
  if( p_usb_dev->descriptor.idVendor != hkp_drv_id_vendor || p_usb_dev->descriptor.idProduct != hkp_drv_id_product )
    //device can't be handled
    //(this is not an error, it's just not our device)
    return NULL;

  //check if high-byte of device-revision id matches
  //(we use high-byte of device-revision to identifiy the device,
  // because all of our devices share the same vendor id and product id)
  if( p_usb_dev->descriptor.bcdDevice >> 8 != hkp_drv_bcd_device_high )
    //device can't be handled
    //(this is not an error, it's just not our device)
    return NULL;

  //now, we are sure, that the device is a hkp_drv-device

  //lock device table
  down( &hkp_drv_dev_table_lock );
  //select a free subminor number
  for( sub_minor = 0; sub_minor < hkp_drv_minor_range; sub_minor++ )
    if( hkp_drv_dev_table[sub_minor] == NULL )
      break;
  //no free subminor found
  if( sub_minor >= hkp_drv_minor_range )
  {
    //print error message
    hkp_drv_error( "too much devices" );
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //device can't be handled
    return NULL;
  }

  //allocate buffer for private data
  p_dev = (struct t_hkp_drv_dev *)kmalloc( sizeof( struct t_hkp_drv_dev ), GFP_KERNEL );
  //error
  if( p_dev == NULL )
  {
    //print error message
    hkp_drv_error( "could not allocate buffer for private data" );
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //abort function
    return NULL;
  }

  //put device into device table
  hkp_drv_dev_table[sub_minor] = p_dev;

  //initialize locks
  init_MUTEX( &p_dev->file_lock );
  init_MUTEX( &p_dev->lock );
  //save parameters of this function
  p_dev->p_usb_dev = p_usb_dev;
  p_dev->interface_no = interface_no;
  p_dev->p_usb_dev_id = p_usb_dev_id;
  //save minor device number
  p_dev->dev_no_sub_minor = sub_minor;
  p_dev->dev_no_minor = hkp_drv_minor_base + p_dev->dev_no_sub_minor;
  //no buffers yet
  p_dev->p_int_out_buffer = NULL;
  p_dev->p_int_in_buffer = NULL;
  p_dev->p_bulk_out_buffer = NULL;
  p_dev->p_bulk_in_buffer = NULL;
  //no URBs yet
  p_dev->p_int_out_urb = NULL;
  p_dev->p_int_in_urb = NULL;
  p_dev->p_bulk_out_urb = NULL;
  p_dev->p_bulk_in_urb = NULL;
  //URBs not submitted/linked yet
  p_dev->int_out_urb_linked = 0;
  p_dev->int_in_urb_linked = 0;
  p_dev->bulk_out_urb_linked = 0;
  p_dev->bulk_in_urb_linked = 0;
  //device is not opened yet
  p_dev->open_cnt = 0;
  //no files have opened this device
  p_dev->file_priv_list.p_dev = NULL;
  p_dev->file_priv_list.p_next = &p_dev->file_priv_list;
  p_dev->file_priv_list.p_prev = &p_dev->file_priv_list;
  //default values for output ports and PWM outputs
  for( i = 0; i < count( p_dev->out_vals ); i++ )
    p_dev->out_vals[i] = 0;
  for( i = 0; i < count( p_dev->pwm_vals ); i++ )
    p_dev->pwm_vals[i] = 0;
  //default values for input ports and analog inputs
  for( i = 0; i < count( p_dev->in_vals ); i++ )
    p_dev->in_vals[i] = 0;
  for( i = 0; i < count( p_dev->ana_vals ); i++ )
    p_dev->ana_vals[i] = 0;

  //get our interface of current configuration
  p_interface = &p_usb_dev->actconfig->interface[interface_no];
  //get interface descriptor
  p_interface_descriptor = &p_interface->altsetting[0];

  //check endpoints: we want: interrupt out, interrupt in, bulk out, bulk in
  endpoints_ok = 0;
  if( p_interface_descriptor->bNumEndpoints == 4 ) //exactly 4 endpoints
  {
    endpoints_ok = 1;
    //interrupt out
    if( (p_interface_descriptor->endpoint[0].bmAttributes & 0x03) == 0x03 //interrupt
      && (p_interface_descriptor->endpoint[0].bEndpointAddress & 0x80) == 0x00 ) //out
    {
      //get interrupt out parameters
      p_dev->int_out_endpoint_addr = p_interface_descriptor->endpoint[0].bEndpointAddress & 0x7F;
      p_dev->int_out_max_packet_size = p_interface_descriptor->endpoint[0].wMaxPacketSize;
      p_dev->int_out_interval = p_interface_descriptor->endpoint[0].bInterval;
    }
    else
    {
      //endpoints are not okay
      endpoints_ok = 0;
    }
    //interrupt in
    if( (p_interface_descriptor->endpoint[1].bmAttributes & 0x03) == 0x03 //interrupt
      && (p_interface_descriptor->endpoint[1].bEndpointAddress & 0x80) != 0x00 ) //in
    {
      //get interrupt in parameters
      p_dev->int_in_endpoint_addr = p_interface_descriptor->endpoint[1].bEndpointAddress & 0x7F;
      p_dev->int_in_max_packet_size = p_interface_descriptor->endpoint[1].wMaxPacketSize;
      p_dev->int_in_interval = p_interface_descriptor->endpoint[1].bInterval;
    }
    else
    {
      //endpoints are not okay
      endpoints_ok = 0;
    }
    //bulk out
    if( (p_interface_descriptor->endpoint[2].bmAttributes & 0x03) == 0x02 //bulk
      && (p_interface_descriptor->endpoint[2].bEndpointAddress & 0x80) == 0x00 ) //out
    {
      //get bulk out parameters
      p_dev->bulk_out_endpoint_addr = p_interface_descriptor->endpoint[2].bEndpointAddress & 0x7F;
      p_dev->bulk_out_max_packet_size = p_interface_descriptor->endpoint[2].wMaxPacketSize;
    }
    else
    {
      //endpoints are not okay
      endpoints_ok = 0;
    }
    //bulk in
    if( (p_interface_descriptor->endpoint[3].bmAttributes & 0x03) == 0x02 //bulk
      && (p_interface_descriptor->endpoint[3].bEndpointAddress & 0x80) != 0x00 ) //in
    {
      //get bulk in parameters
      p_dev->bulk_in_endpoint_addr = p_interface_descriptor->endpoint[3].bEndpointAddress & 0x7F;
      p_dev->bulk_in_max_packet_size = p_interface_descriptor->endpoint[3].wMaxPacketSize;
    }
    else
    {
      //endpoints are not okay
      endpoints_ok = 0;
    }
  }

  //endpoints are not okay
  if( !endpoints_ok )
  {
    //print error message
    hkp_drv_error( "device has got unexpected endpoints" );
    //free device (also frees buffers and URBs)
    hkp_drv_dev_free( p_dev );
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //abort function
    return NULL;
  }

  //allocate buffers
  p_dev->p_int_out_buffer = kmalloc( p_dev->int_out_max_packet_size, GFP_KERNEL );
  p_dev->p_int_in_buffer = kmalloc( p_dev->int_in_max_packet_size, GFP_KERNEL );
  p_dev->p_bulk_out_buffer = kmalloc( p_dev->bulk_out_max_packet_size, GFP_KERNEL );
  p_dev->p_bulk_in_buffer = kmalloc( p_dev->bulk_in_max_packet_size, GFP_KERNEL );
  if( p_dev->p_int_out_buffer == NULL || p_dev->p_int_in_buffer == NULL
   || p_dev->p_bulk_out_buffer == NULL || p_dev->p_bulk_in_buffer == NULL )
  {
    //print error message
    hkp_drv_error( "could not allocate buffers" );
    //free device (also frees buffers and URBs)
    hkp_drv_dev_free( p_dev );
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //abort function
    return NULL;
  }

  //allocate URBs
  p_dev->p_int_out_urb = usb_alloc_urb( 0 );
  p_dev->p_int_in_urb = usb_alloc_urb( 0 );
  p_dev->p_bulk_out_urb = usb_alloc_urb( 0 );
  p_dev->p_bulk_in_urb = usb_alloc_urb( 0 );
  if( p_dev->p_int_out_urb == NULL || p_dev->p_int_in_urb == NULL
   || p_dev->p_bulk_out_urb == NULL || p_dev->p_bulk_in_urb == NULL )
  {
    //print error message
    hkp_drv_error( "could not allocate URBs" );
    //free device (also frees buffers and URBs)
    hkp_drv_dev_free( p_dev );
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //abort function
    return NULL;
  }

  //set up data in buffer for interrupt out
  memset( p_dev->p_int_out_buffer, 0, p_dev->int_out_max_packet_size );
  //data for output ports
  if( p_dev->int_out_max_packet_size >= 1 )
    p_dev->p_int_out_buffer[0] = p_dev->out_vals[0];
  if( p_dev->int_out_max_packet_size >= 2 )
    p_dev->p_int_out_buffer[1] = p_dev->out_vals[1];
  //data for PWM outputs
  if( p_dev->int_out_max_packet_size >= 4 )
    *(__u16 *)&p_dev->p_int_out_buffer[2] = __cpu_to_be16( (p_dev->pwm_vals[0] & 0x3FF) << 6 );
  if( p_dev->int_out_max_packet_size >= 6 )
    *(__u16 *)&p_dev->p_int_out_buffer[4] = __cpu_to_be16( (p_dev->pwm_vals[1] & 0x3FF) << 6 );
  //set up data in interrupt out URB
  FILL_INT_URB( p_dev->p_int_out_urb, p_usb_dev, usb_sndintpipe( p_usb_dev, p_dev->int_out_endpoint_addr ),
      p_dev->p_int_out_buffer, p_dev->int_out_max_packet_size, hkp_drv_int_out_callback,
      p_dev, p_dev->int_out_interval );
  //interrupt out URB will be linked/submitted soon
  //(do this here because callback could perhaps be called from usb_submit_urb)
  p_dev->int_out_urb_linked = 1;
  //submit interrupt out URB
  if( usb_submit_urb( p_dev->p_int_out_urb ) )
  {
    //interrupt out URB is not linked/submitted
    p_dev->int_out_urb_linked = 0;
    //print error message
    hkp_drv_error( "could not submit interrupt out URB" );
    //free device (also frees buffers and URBs)
    hkp_drv_dev_free( p_dev );
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //abort function
    return NULL;
  }
  //print info that interrupt in was started
  hkp_drv_debugf( "interrupt input started (max_len=%d)", p_dev->p_int_in_urb->transfer_buffer_length );

  //set up data in interrupt in URB
  FILL_INT_URB( p_dev->p_int_in_urb, p_usb_dev, usb_rcvintpipe( p_usb_dev, p_dev->int_in_endpoint_addr ),
      p_dev->p_int_in_buffer, p_dev->int_in_max_packet_size, hkp_drv_int_in_callback,
      p_dev, p_dev->int_in_interval );
  //interrupt in URB will be linked/submitted soon
  //(do this here because callback could perhaps be called from usb_submit_urb)
  p_dev->int_in_urb_linked = 1;
  //submit interrupt in URB
  if( usb_submit_urb( p_dev->p_int_in_urb ) )
  {
    //interrupt in URB is not linked/submitted
    p_dev->int_in_urb_linked = 0;
    //print error message
    hkp_drv_error( "could not submit interrupt in URB" );
    //free device (also frees buffers and URBs)
    hkp_drv_dev_free( p_dev );
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //abort function
    return NULL;
  }
  //print info that interrupt in was started
  hkp_drv_debugf( "interrupt input started (max_len=%d)", p_dev->p_int_in_urb->transfer_buffer_length );

  //set up data in bulk in URB
  FILL_BULK_URB( p_dev->p_bulk_in_urb, p_usb_dev, usb_rcvbulkpipe( p_usb_dev, p_dev->bulk_in_endpoint_addr ),
      p_dev->p_bulk_in_buffer, p_dev->bulk_in_max_packet_size, hkp_drv_bulk_in_callback,
      p_dev );
  //bulk in URB will be linked/submitted soon
  //(do this here because callback could perhaps be called from usb_submit_urb)
  p_dev->bulk_in_urb_linked = 1;
  //submit bulk in URB
  if( usb_submit_urb( p_dev->p_bulk_in_urb ) )
  {
    //bulk in URB is not linked/submitted
    p_dev->bulk_in_urb_linked = 0;
    //print error message
    hkp_drv_error( "could not submit bulk in URB" );
    //free device (also frees buffers and URBs)
    hkp_drv_dev_free( p_dev );
    //unlock device table
    up( &hkp_drv_dev_table_lock );
    //abort function
    return NULL;
  }
  //print info that bulk in was started
  hkp_drv_debugf( "bulk input started (max_len=%d)", p_dev->p_bulk_in_urb->transfer_buffer_length );

  //register devfs node
  snprintf( p_dev->devfs_node_name, sizeof( p_dev->devfs_node_name ), hkp_drv_devfs_node_name"%d", p_dev->dev_no_sub_minor );
  p_dev->h_devfs_node = devfs_register( usb_devfs_handle, p_dev->devfs_node_name, DEVFS_FL_DEFAULT, USB_MAJOR, p_dev->dev_no_minor,
    S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH, &hkp_drv_file_operations, NULL );

  //print info message
  hkp_drv_infof( "device attached (major %d, minor %d, devfs name %s)", USB_MAJOR, p_dev->dev_no_minor, p_dev->devfs_node_name );

  //unlock device table
  up( &hkp_drv_dev_table_lock );

  //return pointer to private data
  return (void *)p_dev;
}

//device is removed from USB
static void hkp_drv_disconnect( struct usb_device * p_usb_dev, void * p_device )
{
  struct t_hkp_drv_dev * p_dev;

  //get pointer to device
  p_dev = (struct t_hkp_drv_dev *)p_device;

  //lock device table and device
  down( &hkp_drv_dev_table_lock );
  down( &p_dev->lock );

  //print info message
  hkp_drv_infof( "device disconnected (major %d, minor %d, devfs name %s)", USB_MAJOR, p_dev->dev_no_minor, p_dev->devfs_node_name );

  //unregister devfs node
  p_dev->devfs_node_name[0] = 0;
  devfs_unregister( p_dev->h_devfs_node );

  //device is opened by someone
  if( p_dev->open_cnt > 0 )
  {
    //remove pointer to usb device
    //(device is gone)
    p_dev->p_usb_dev = NULL;
    //unlock the device
    up( &p_dev->lock );
  }
  //device is not opened by anyone
  else
  {
    //unlock the device
    up( &p_dev->lock );
    //free the device (and remove it from the device table)
    hkp_drv_dev_free( p_device );
  }

  //unlock device table
  up( &hkp_drv_dev_table_lock );
}

//table of supported devices
static struct usb_device_id hkp_drv_id_table [] =
{
  { USB_DEVICE( hkp_drv_id_vendor, hkp_drv_id_product ) },
  { } //end of table
};

//structure for USB data and functions
static struct usb_driver hkp_drv_usb_driver =
{
  name:        "hkp_drv",
  probe:       hkp_drv_probe,
  disconnect:  hkp_drv_disconnect,
  fops:        &hkp_drv_file_operations,
  minor:       hkp_drv_minor_base,
  id_table:    hkp_drv_id_table,
};



//*** module init and exit

//module is being loaded
static int hkp_drv_init( )
{
  //local variables
  int i;
  
#ifdef DEBUG_MEM
  //initialize memory debug lock
  init_MUTEX( &my_alloc_lock );
#endif
  
  //short message
  hkp_drv_info( hkp_drv_name" ("hkp_drv_ver")" );

  //initialize table with pointers to private device data of devices
  for( i = 0; i < hkp_drv_minor_range; i++ )
    hkp_drv_dev_table[i] = NULL;
  //initialize the lock for this table
  init_MUTEX( &hkp_drv_dev_table_lock );

  //reset boolean flags for tasks to do in next hkp_drv_task
  //(hkp_drv_task is task for things that cannot be done in interrupt)
  //(use test_and_x to be smp safe)
  for( i = 0; i < hkp_drv_minor_range; i++ )
    test_and_clear_bit( hkp_drv_task_flag_bulk_out_try, &hkp_drv_task_flags[i] );

  //register this driver with the USB subsystem
  i = usb_register( &hkp_drv_usb_driver );
  //an error occured
  if( i < 0 )
  {
    //error-message
    hkp_drv_errorf( "usb_register failed with %d", i );
    //init failed
    return -1;
  }

  //init succeeded
  hkp_drv_info( "loaded..." );
  return 0;
}

//module is being unloaded
static void hkp_drv_exit( )
{
  //deregister driver with the USB subsystem
  usb_deregister( &hkp_drv_usb_driver );

  //exit done
  hkp_drv_info( "unloaded..." );
}



//module init and exit procedure
module_init( hkp_drv_init );
module_exit( hkp_drv_exit );

//table with supported devices
MODULE_DEVICE_TABLE( usb, hkp_drv_id_table );

